url = "http://localhost:6363/"
json_context = {
    "_": "_:",
    "doc": "terminusdb:///system/data/",
    "layer": "http://terminusdb.com/schema/layer#",
    "owl": "http://www.w3.org/2002/07/owl#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "ref": "http://terminusdb.com/schema/ref#",
    "repo": "http://terminusdb.com/schema/repository#",
    "terminus": "http://terminusdb.com/schema/terminus#",
    "vio": "http://terminusdb.com/schema/vio#",
    "woql": "http://terminusdb.com/schema/woql#",
    "xdd": "http://terminusdb.com/schema/xdd#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
}

dbrec = {
    "db": "aaaaaa",
    "account": "admin",
    "title": "admin|aaaaaa",
    "description": "dasdasdds",
}

dbrecs = [
    {"db": "5534534", "account": "admin", "title": "3453", "description": "345345"},
    {
        "db": "adsasasddsa",
        "account": "admin",
        "title": "admin|adsasasddsa",
        "description": "asdsadsda",
    },
    {"db": "blah", "account": "admin", "title": "admin|blah", "description": "adfadf"},
    {
        "db": "daassd",
        "account": "admin",
        "title": "admin|daassd",
        "description": "asdasdsd",
    },
    {
        "db": "dassadds",
        "account": "admin",
        "title": "admin|dassadds",
        "description": "asdsdsad",
    },
    {"db": "ddd", "account": "admin", "title": "admin|ddd", "description": "ddds"},
    {"db": "ffff", "account": "admin", "title": "admin|ffff", "description": "fffff"},
    {
        "db": "fffffff",
        "account": "admin",
        "title": "admin|fffffff",
        "description": "fff",
    },
    {
        "db": "ggggggggg",
        "account": "admin",
        "title": "admin|ggggggggg",
        "description": "gg",
    },
    {
        "db": "twretwert",
        "account": "admin",
        "title": "admin|twretwert",
        "description": "adfadf",
    },
    {
        "db": "terminus",
        "account": "",
        "title": "Master Database",
        "description": "The master database contains the meta-data about databases, users and roles",
    },
]



import secrets
import string

from terminusdb_client.woqlclient.connectionConfig import ConnectionConfig


class TestConnectionConfig:
    start_server_url = "http://localhost:6363/"
    start_dbid = "testDB"
    local_user = "admin"
    # Set of random string values that can as replacements for dummy id valuess
    # i.e accountid,branchid,refid etc
    id_value = "".join(
        secrets.choice(string.ascii_uppercase + string.ascii_lowercase)
        for _ in range(16)
    )
    # to be review !!!!!
    connection_config = ConnectionConfig(
        start_server_url,
        db=start_dbid,
        user=local_user,
        key="mykey",
        account=local_user,
    )

    db_url = "http://localhost:6363/api/db/admin/testDB"

    def test_get_sever_url(self):
        assert self.connection_config.server == self.start_server_url
        assert self.connection_config.db_url() == self.db_url

    def test_change_branch(self):
        self.connection_config.branch = "myBranch"
        assert self.connection_config.db_url() == self.db_url
        assert (
            self.connection_config.query_url()
            == "http://localhost:6363/api/woql/admin/testDB/local/branch/myBranch"
        )

    def test_change_server(self):
        new_ref = "gfhfjkflfgorpyuiioo"
        self.connection_config.ref = new_ref
        assert (
            self.connection_config.query_url()
            == "http://localhost:6363/api/woql/admin/testDB/local/commit/" + new_ref
        )

    def test_check_basic_auth(self):
        assert self.connection_config.basic_auth == "admin:mykey"

    def test_set_remote_auth(self):
        auth_dict = {type: "jwt", "key": "eyJhbGciOiJIUzI1NiIsInR5c"}
        self.connection_config.set_remote_auth(auth_dict)
        assert self.connection_config.remote_auth == auth_dict
        auth_dict = {
            type: "basic",
            "user": self.local_user,
            "key": "admin_testDB_Password",
        }
        self.connection_config.update(remote_auth=auth_dict)
        assert self.connection_config.remote_auth == auth_dict

    def test_update(self):
        self.connection_config.update(db=self.id_value)
        self.connection_config.update(account=self.id_value)
        self.connection_config.update(repo=self.id_value)
        self.connection_config.update(branch=self.id_value)
        self.connection_config.update(ref=self.id_value)
        self.connection_config.update(repo=self.id_value)
        assert self.connection_config.db == self.id_value
        assert self.connection_config.account == self.id_value
        assert self.connection_config.repo == self.id_value
        assert self.connection_config.branch == self.id_value
        assert self.connection_config.ref == self.id_value
        assert self.connection_config.repo == self.id_value

    def test_clear_cursor(self):
        self.connection_config.update(branch=self.id_value)
        self.connection_config.update(repo=self.id_value)
        self.connection_config.update(account=self.id_value)
        self.connection_config.update(db=self.id_value)
        self.connection_config.update(ref=self.id_value)
        self.connection_config.clear_cursor()
        assert self.connection_config.branch != self.id_value
        assert self.connection_config.repo != self.id_value
        assert self.connection_config.account is False
        assert self.connection_config.db is False
        assert self.connection_config.ref is False

    def test_user(self):
        assert self.connection_config.user() == self.local_user
        auth_dict = {type: "jwt", "key": "eyJhbGciOiJIUzI1NiIsInR5c"}
        self.connection_config.update(remote=auth_dict)
        assert self.connection_config.user() == self.local_user

    def test_set_basic_auth(self):
        assert self.connection_config.set_basic_auth() is None
        self.connection_config.set_basic_auth(api_key="admin_testDB_Password")
        basic_auth = self.connection_config.basic_auth
        assert basic_auth == "admin:admin_testDB_Password"

    def test_db_url_fragmen(self):
        self.connection_config.update(db=self.id_value)
        self.connection_config.update(account=self.id_value)
        db_url_fragment = self.connection_config.db_url_fragment()
        assert db_url_fragment == f"{self.id_value}/{self.id_value}"

    def test_db_base(self):
        self.connection_config.update(db=self.id_value)
        self.connection_config.update(account=self.id_value)
        db_base = self.connection_config.db_base("push")
        assert (
            db_base == f"http://localhost:6363/api/push/{self.id_value}/{self.id_value}"
        )

    def test_repo_base(self):
        self.connection_config.update(db=self.id_value)
        self.connection_config.update(account=self.id_value)
        self.connection_config.db_base("pull")
        default_base = self.connection_config.repo_base("pull")
        assert (
            default_base
            == f"http://localhost:6363/api/pull/{self.id_value}/{self.id_value}/local"
        )

        self.connection_config.update(repo=self.id_value)
        updated_base = self.connection_config.db_base("pull")
        assert (
            updated_base
            == f"http://localhost:6363/api/pull/{self.id_value}/{self.id_value}"
        )

    def test_account(self):
        self.connection_config.account = self.id_value
        assert self.connection_config.account == self.id_value

    def test_db(self):
        self.connection_config.db = self.id_value
        assert self.connection_config.db == self.id_value

    def test_repo(self):
        self.connection_config.repo = self.id_value
        assert self.connection_config.repo == self.id_value

    def branch(self):
        self.connection_config.branch = self.id_value
        assert self.connection_config.branch == self.id_value

    def ref(self):
        self.connection_config.ref = self.id_value
        assert self.connection_config.ref == self.id_value



# from .errorMessage import ErrorMessage
import json
import warnings
from base64 import b64encode

import requests
import terminusdb_client.woql_utils as utils
from urllib3.exceptions import InsecureRequestWarning

from .api_endpoint_const import APIEndpointConst
from .errors import APIError


def _verify_check(url, insecure=False):
    if url[:17] == "https://127.0.0.1" or url[:7] == "http://" or insecure:
        return False
    else:
        return True


class DispatchRequest:
    def __init__(self):
        pass

    @staticmethod
    def __get_call(url, headers, payload, insecure=False):
        url = utils.add_params_to_url(url, payload)
        if not _verify_check(url, insecure):
            warnings.simplefilter("ignore", InsecureRequestWarning)
        result = requests.get(url, headers=headers, verify=_verify_check(url, insecure))
        warnings.resetwarnings()
        return result

    @staticmethod
    def __post_call(url, headers, payload, file_dict=None, insecure=False):
        if not _verify_check(url, insecure):
            warnings.simplefilter("ignore", InsecureRequestWarning)
        if file_dict:
            file_dict["payload"] = (
                "payload",
                json.dumps(payload),
                "application/json",
            )

            result = requests.post(
                url,
                headers=headers,
                files=file_dict,
                verify=_verify_check(url, insecure),
            )
            # Close the files although request should do this :(
            for key in file_dict:
                (_, stream, _) = file_dict[key]
                if type(stream) != str:
                    stream.close()
        else:
            headers["content-type"] = "application/json"
            result = requests.post(
                url, json=payload, headers=headers, verify=_verify_check(url, insecure)
            )
        warnings.resetwarnings()
        return result

    @staticmethod
    def __put_call(url, headers, payload, file_dict=None, insecure=None):
        if not _verify_check(url):
            warnings.simplefilter("ignore", InsecureRequestWarning)
        if file_dict:
            file_dict["payload"] = (
                "payload",
                json.dumps(payload),
                "application/json",
            )

            result = requests.put(
                url, headers=headers, files=file_dict, verify=_verify_check(url),
            )
            # Close the files although request should do this :(
            for key in file_dict:
                (_, stream, _) = file_dict[key]
                if type(stream) != str:
                    stream.close()
        else:
            headers["content-type"] = "application/json"
            result = requests.put(
                url, json=payload, headers=headers, verify=_verify_check(url, insecure)
            )
        warnings.resetwarnings()
        return result

    @staticmethod
    def __delete_call(url, headers, payload, insecure=False):
        url = utils.add_params_to_url(url, payload)
        if not _verify_check(url, insecure):
            warnings.simplefilter("ignore", InsecureRequestWarning)
        result = requests.delete(
            url, headers=headers, verify=_verify_check(url, insecure)
        )
        warnings.resetwarnings()
        return result

    @staticmethod
    def __autorization_header(basic_auth=None, remote_auth=None):
        headers = {}

        # if (payload and ('terminus:user_key' in  payload)):
        # utils.encodeURIComponent(payload['terminus:user_key'])}
        if basic_auth:
            headers["Authorization"] = "Basic %s" % b64encode(
                (basic_auth).encode("utf-8")
            ).decode("utf-8")
        if remote_auth and remote_auth["type"] == "jwt":
            headers["Authorization-Remote"] = "Bearer %s" % remote_auth["key"]
        elif remote_auth and remote_auth["type"] == "basic":
            rauthstr = remote_auth["user"] + ":" + remote_auth["key"]
            headers["Authorization-Remote"] = "Basic %s" % b64encode(
                (rauthstr).encode("utf-8")
            ).decode("utf-8")
        return headers

    # url, action, payload, basic_auth, jwt=null

    @classmethod
    def send_request_by_action(
        cls,
        url,
        action,
        payload=None,
        basic_auth=None,
        remote_auth=None,
        file_dict=None,
        insecure=False,
    ):

        # payload default as empty dict is against PEP
        # print("Sending to URL____________", url)
        # print("Send Request By Action_____________", action)
        if payload is None:
            payload = {}

        try:
            request_response = None
            headers = cls.__autorization_header(basic_auth, remote_auth)

            if action in [
                APIEndpointConst.GET_TRIPLES,
                APIEndpointConst.GET_CSV,
                APIEndpointConst.CONNECT,
                APIEndpointConst.CLASS_FRAME,
            ]:
                request_response = cls.__get_call(
                    url, headers, payload, insecure=insecure
                )

            elif action in [
                APIEndpointConst.DELETE_DATABASE,
                APIEndpointConst.DELETE_GRAPH,
            ]:
                request_response = cls.__delete_call(
                    url, headers, payload, insecure=insecure
                )

            elif action in [
                APIEndpointConst.WOQL_QUERY,
                APIEndpointConst.CREATE_DATABASE,
                APIEndpointConst.UPDATE_TRIPLES,
                APIEndpointConst.UPDATE_CSV,
                APIEndpointConst.CREATE_GRAPH,
                APIEndpointConst.FETCH,
                APIEndpointConst.PULL,
                APIEndpointConst.PUSH,
                APIEndpointConst.REBASE,
                APIEndpointConst.BRANCH,
                APIEndpointConst.CLONE,
                APIEndpointConst.RESET,
                APIEndpointConst.OPTIMIZE,
                APIEndpointConst.SQUASH,
            ]:
                request_response = cls.__post_call(
                    url, headers, payload, file_dict, insecure=insecure
                )

            elif action in [
                APIEndpointConst.INSERT_TRIPLES,
                APIEndpointConst.INSERT_CSV,
            ]:

                request_response = cls.__put_call(
                    url, headers, payload, file_dict, insecure=insecure
                )

            if request_response.status_code == 200:
                # print("hellow ")
                return request_response.json()  # if not a json not it raises an error
            else:
                # Raise an exception if a request is unsuccessful
                message = "Api Error"

                if type(request_response.text) is str:
                    message = request_response.text

                raise (
                    APIError(
                        message,
                        url,
                        request_response.json(),
                        request_response.status_code,
                    )
                )

        # to be reviewed
        # the server in the response return always content-type application/json
        except ValueError:
            # if the response type is not a json
            return request_response

        """
        except Exception as err:
            print(type(err))
            print(err.args)

        except requests.exceptions.RequestException as err:
            print ("Request Error",err)
        except requests.exceptions.HTTPError as err:
            print ("Http Error:",err)
        except requests.exceptions.ConnectionError as err:
            print ("Error Connecting:",err)
        except requests.exceptions.Timeout as err:
            print ("Timeout Error:",err)
        """



import pytest


@pytest.fixture(scope="module")
def one_class_obj():
    return {
        "@type": "api:WoqlResponse",
        "api:status": "api:success",
        "api:variable_names": [
            "Class ID",
            "Class Name",
            "Description",
            "Parents",
            "Children",
            "Abstract",
        ],
        "bindings": [
            {
                "Abstract": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "No",
                },
                "Children": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "",
                },
                "Class ID": "terminusdb:///schema#Journey",
                "Class Name": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "Bike Journey",
                },
                "Description": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "Bike Journey object that capture each bike joourney.",
                },
                "Parents": ["http://terminusdb.com/schema/system#Document"],
            }
        ],
        "deletes": 0,
        "inserts": 0,
        "transaction_retry_count": 0,
    }


@pytest.fixture(scope="module")
def one_class_prop():
    return {
        "@type": "api:WoqlResponse",
        "api:status": "api:success",
        "api:variable_names": [
            "Property ID",
            "Property Name",
            "Property Domain",
            "Property Type",
            "Property Range",
            "Property Description",
        ],
        "bindings": [
            {
                "Property Description": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "",
                },
                "Property Domain": "terminusdb:///schema#Journey",
                "Property ID": "terminusdb:///schema#Duration",
                "Property Name": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "",
                },
                "Property Range": "http://www.w3.org/2001/XMLSchema#integer",
                "Property Type": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "Data",
                },
            }
        ],
        "deletes": 0,
        "inserts": 0,
        "transaction_retry_count": 0,
    }


@pytest.fixture(scope="module")
def one_object():
    return {
        "@type": "api:WoqlResponse",
        "api:status": "api:success",
        "api:variable_names": ["Object ID", "Object Type"],
        "bindings": [
            {
                "Object ID": "terminusdb:///data/Journey_myobj",
                "Object Type": "terminusdb:///schema#Journey",
            }
        ],
        "deletes": 0,
        "inserts": 0,
        "transaction_retry_count": 0,
    }


@pytest.fixture(scope="module")
def one_prop_val():
    return {
        "@type": "api:WoqlResponse",
        "api:status": "api:success",
        "api:variable_names": [
            "Object ID",
            "Property ID",
            "Property Value",
            "Value ID",
            "Value Class",
        ],
        "bindings": [
            {
                "Object ID": "terminusdb:///data/Journey_myobj",
                "Property ID": "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
                "Property Value": "terminusdb:///schema#Journey",
                "Value Class": "system:unknown",
                "Value ID": "system:unknown",
            },
            {
                "Object ID": "terminusdb:///data/Journey_myobj",
                "Property ID": "terminusdb:///schema#Duration",
                "Property Value": {
                    "@type": "http://www.w3.org/2001/XMLSchema#integer",
                    "@value": 75,
                },
                "Value Class": "system:unknown",
                "Value ID": "system:unknown",
            },
        ],
        "deletes": 0,
        "inserts": 0,
        "transaction_retry_count": 0,
    }



# connectionCapabilities.py


# const UTILS = require('./utils.js');

import pprint

pp = pprint.PrettyPrinter(indent=4)

"""
    Creates an entry in the connection registry for the server
    and all the databases that the client has access to
    maps the input authorties to a per-db array for internal storage and easy
    access control checks
    {doc:dbid => {system:authority =>
    [system:woql_select, system:create_document, auth3, ...]}}
"""


class ConnectionCapabilities:
    def __init__(self):
        self.user = None
        self.dbdocs = {}
        self.orgdocs = {}
        self.databases = None
        self._jsonld_context = {}
        self._systemdb_context = {}

    def _clear(self):
        self.user = None
        self.dbdocs = {}
        self.orgdocs = {}
        self._jsonld_context = {}
        self._systemdb_context = {}

    def set_capabilities(self, capabilities=None):
        self._clear()
        self.databases = None
        if capabilities is not None:
            ctxt = capabilities.get("@context")
            if ctxt is not None:
                self._load_connection_context(ctxt)
        self.user = self._extract_user_info(capabilities)
        return self._extract_database_organizations()

    def get_databases(self):
        if self.databases is None:
            self.databases = self._databases_from_dbdocs()
        return self.databases

    def set_databases(self, newdbs):
        self.databases = newdbs

    def get_database(self, dbid, orgid):
        if self.databases is None:
            self.databases = self._databases_from_dbdocs()
        for db in self.databases:
            if db["id"] == dbid and db["organization"] == orgid:
                return db
        return None

    def get_user(self):
        return self.user

    def author(self):
        if self.user.get("author"):
            return self.user.get("author")
        else:
            return self.user.get("id")

    def _databases_from_dbdocs(self):
        dbs = []
        for docid in self.dbdocs:
            if not self.dbdocs[docid]["system"]:
                dbs.append(self._get_db_rec(self.dbdocs[docid]))
        return dbs

    def set_roles(self, role_output):
        cntnt = role_output.get("bindings")
        if cntnt is None:
            return False
        for pred in cntnt:
            onerec = {}
            onerec["dbdocid"] = pred["Database_ID"]
            onerec["organization"] = pred["Organization"]["@value"]
            roles = self._multiple_rdf_objects(pred["Owner_Role_Obj"], "system:Role")
            self.dbdocs[onerec["dbdocid"]] = roles
        self._extract_database_organizations()

        rdb = role_output.get("databases")
        if rdb:
            self.set_databases(rdb)
        odb = role_output.get("organizations")
        if odb:
            self.set_organizations(odb)

    def get_organizations(self):
        if self.organizations is None:
            self.organizations = self._organizations_from_orgdocs()
        return self.organizations

    def set_organizations(self, newdbs):
        self.organizations = newdbs

    def get_organization(self, orgid):
        if self.organizations is None:
            self.organizations = self._organizations_from_orgdocs()
        for org in self.organizations:
            if org["id"] == orgid:
                return org
        return None

    def _organizations_from_orgdocs(self):
        return self.orgdocs.values()

    def actions_permitted(self, actions, resnames):
        actlist = self._toarr(actions)
        reslist = self._toarr(resnames)
        for i in actlist.keys():
            for j in reslist.keys():
                if self._roles_cover_resource_action(actlist[i], reslist[j]):
                    return False
        return True

    def get_context_for_outbound_query(self, woql, dbid):
        """TODO: This should be refactored later"""
        return {}

    def _is_system_db(self, dbid):
        if dbid == "_system":
            return True
        return False

    def get_json_context(self):
        return self._jsonld_context

    def set_json_context(self, ctxt):
        self._jsonld_context = ctxt

    def get_system_context(self):
        return self._systemdb_context

    def _roles_cover_resource_action(self, action, resname):
        for docid in self.user["roles"].keys():
            if self._role_covers_resource_action(
                self.user["roles"][docid], action, resname
            ):
                return True
        return False

    def _role_covers_resource_action(self, role, action, resname):
        for docid in role["capabilities"].keys():
            if self._capability_covers_resource_action(
                role["capabilities"][docid], action, resname
            ):
                return True
        return False

    def _capability_covers_resource_action(self, cap, action, resname):
        if action[0:7] != "system:":
            action = "system:" + action
        if action not in cap["actions"]:
            return False
        if resname not in cap["resources"]:
            return False
        return True

    def _extract_user_info(self, capabilities):
        info = self._extract_rdf_basics(capabilities)
        info["id"] = self._single_rdf_value("system:agent_name", capabilities)
        info["author"] = self._single_rdf_value("system:user_identifier", capabilities)
        croles = capabilities.get("system:role")
        if croles is not None:
            info["roles"] = self._multiple_rdf_objects(croles, "system:Role")
        else:
            info["roles"] = []
        return info

    def _extract_user_role(self, jrole):
        nrole = self._extract_rdf_basics(jrole)
        croles = jrole.get("system:capability")
        if croles is not None:
            nrole["capabilities"] = self._multiple_rdf_objects(
                croles, "system:Capability"
            )
        else:
            nrole["capabilities"] = []
        return nrole

    def _extract_role_capability(self, jcap):
        nrole = self._extract_rdf_basics(jcap)
        acts = jcap.get("system:action")
        resources = jcap.get("system:capability_scope")
        if acts is not None:
            nrole["actions"] = self._extract_multiple_ids(acts)
        if resources is not None:
            nrole["resources"] = self._extract_capability_resources(resources)
        return nrole

    def _extract_capability_resources(self, scope):
        rnames = []
        resources = self._toarr(scope)
        for res in resources:
            rnames.append(self._extract_resource_name(res))
            rid = res.get("@id")
            rtype = res.get("@type")
            if not (rid is None or rtype is None):
                exist = self.dbdocs.get(rid)
                if exist is None and (
                    rtype == "system:SystemDatabase" or rtype == "system:Database"
                ):
                    self.dbdocs[rid] = self._extract_database(res)
                if rtype == "system:Organization":
                    self.orgdocs[rid] = self._extract_organization(res)
        return rnames

    def _extract_resource_name(self, jres):
        sys_id = self._single_rdf_value("system:resource_name", jres)
        if sys_id is None:
            sys_id = self._single_rdf_value("system:organization_name", jres)
        if sys_id is None:
            sys_id = self._single_rdf_value("system:database_name", jres)
        return sys_id

    def _extract_database(self, jres):
        db = self._extract_rdf_basics(jres)
        db["id"] = self._extract_resource_name(jres)
        if jres.get("@type") == "system:SystemDatabase":
            db["system"] = True
        else:
            db["system"] = False
        return db

    def _get_db_rec(self, rec):
        urec = {}
        for pred in rec.keys():
            if pred != "system":
                urec[pred] = rec[pred]
        return urec

    def _extract_database_organizations(self):
        for docid in self.dbdocs:
            for odocid in self.orgdocs:
                dbs = self.orgdocs[odocid].get("databases")
                if dbs is not None and docid in dbs:
                    self.dbdocs[docid]["organization"] = self.orgdocs[odocid]["id"]

    def _extract_organization(self, jres):
        org = self._extract_rdf_basics(jres)
        org["id"] = self._extract_resource_name(jres)
        dbs = jres.get("system:organization_database")
        kids = jres.get("system:organization_child")
        incs = jres.get("system:resource_includes")
        if dbs is not None:
            org["databases"] = self._extract_multiple_ids(dbs)
        if kids is not None:
            org["children"] = self._extract_multiple_ids(kids)
        if incs is not None:
            org["includes"] = self._extract_multiple_ids(incs)
        return org

    def _extract_multiple_ids(self, jres):
        jids = self._toarr(jres)
        ids = []
        for k in jids:
            ids.append(k["@id"])
        return ids

    def _multiple_rdf_objects(self, rdf, con_type):
        nvals = {}
        vals = self._toarr(rdf)
        for n in vals:
            if isinstance(n, dict):
                nid = n.get("@id")
                if nid is not None:
                    nvals[nid] = self._extract_rdf_object(con_type, n)
        return nvals

    def _extract_rdf_object(self, con_type, jres):
        if con_type == "system:Role":
            return self._extract_user_role(jres)
        elif con_type == "system:Capability":
            return self._extract_role_capability(jres)

    def _extract_rdf_basics(self, jres):
        info = {}
        info["label"] = self._single_rdf_value("rdfs:label", jres)
        info["comment"] = self._single_rdf_value("rdfs:comment", jres)
        return info

    def _single_rdf_value(self, pred, rdf):
        if rdf.get(pred) is None:
            return ""
        if isinstance(rdf[pred], list):
            return rdf[pred][0]["@value"]
        return rdf[pred]["@value"]

    def _multiple_rdf_values(self, pred, rdf):
        vals = []
        if rdf.get(pred) is None:
            return vals
        rdflist = self._toarr(rdf.get(pred))
        for item in rdflist:
            vals.append(item["@value"])
        return vals

    def _toarr(self, el):
        if isinstance(el, list):
            return el
        return [el]

    def _load_connection_context(self, ctxt):
        self._systemdb_context = ctxt
        for k in ctxt.keys():
            if k != "doc":
                self._jsonld_context[k] = ctxt[k]



from datetime import datetime
from typing import Dict, List, Union
from urllib.parse import quote

from .woql_library import WOQLLib
from .woql_query import WOQLQuery

WOQLTYPE_TO_PYTYPE = {
    "string": str,
    "boolean": bool,
    "integer": int,
    "decimal": float,
    "dateTime": datetime,
}


class WOQLClass:
    def __init__(
        self,
        obj_id: str,
        label: str = None,
        description: str = None,
        obj_property: dict = None,
    ):
        """The WOQLClass constructor.

        Parameters
        ----------
        obj_id: str
            Object id for which the instance is created.
        label: str, optional
            Optional label to be given for the instance.
        description: str, optional
            Optional description to be given for the instance.
        obj_property: dict, optional
            Optional properties which can be provided during initialization.

        Examples
        --------
        >>> woql_object_prop_des = WOQLClass(
                "Journey",
                label="Journey Object",
                description="A car Journey object."
                obj_property={
                    "Duration": {
                        "type": "dateTime",
                        "label": "Journey Duration",
                        "description": "Journey duration in minutes.",
                    }
                },
            )
        """
        self.id = obj_id
        self._label = label
        self._description = description
        if obj_property is not None:
            self._property = obj_property
        else:
            self._property = {}

        self.query_obj = WOQLQuery().doctype(
            obj_id, label=label, description=description
        )
        if obj_property is not None:
            for key, item in obj_property.items():
                self.query_obj = self.query_obj.property(
                    key, item.get("type"), item.get("label"), item.get("description")
                )

    def __str__(self):
        """Returns the id of the instance

        Returns
        -------
        str

        """
        return self.id

    def __eq__(self, other):
        """To check if the object is equal to the given object.

        Parameters
        ----------
        other: WOQLClass
            The other object which has to be compared with.

        Returns
        -------
        bool

        """
        if isinstance(other, WOQLClass):
            return self.id == other.id
        else:
            return False

    @property
    def label(self) -> str:
        """Returns the label of the instance.

        Returns
        -------
        str

        """
        if self._label is None:
            return ""
        return self._label

    @label.setter
    def label(self, label: str):
        """Sets the label for this instance.

        Parameters
        ----------
        label: str
            Label which needs to be set.

        """
        self.query_obj = self.query_obj.label(label)
        self._label = label

    @property
    def description(self) -> str:
        """Returns the description of the instance.

        Returns
        -------
        str

        """
        if self._description is None:
            return ""
        return self._description

    @description.setter
    def description(self, description: str):
        """Sets the description for this instance.

        Parameters
        ----------
        description: str
            description which needs to be set.
        """
        self.query_obj = self.query_obj.description(description)
        self._description = description

    def add_property(
        self,
        pro_id: str,
        property_type: Union[str, "WOQLClass"],
        label: str = None,
        description: str = None,
    ):
        """Adds the given properties to the instance.

        Parameters
        ----------
        pro_id: str
            Property id
        property_type: str (or) WOQLClass
            Properties to be given in json or as WOQLClass instance.
        label: str, optional
            Optional label to be given for the instance.
        description: str, optional
            Optional description to be given for the instance.

        Returns
        -------
        WOQLClass
        """
        if isinstance(property_type, str):
            self.query_obj = self.query_obj.property(
                pro_id, property_type, label, description
            )
        elif isinstance(property_type, WOQLClass):
            self.query_obj = self.query_obj.property(
                pro_id, property_type.id, label, description
            )
        else:
            raise ValueError("property_type needs to be either string or WOQLClass")
        self._property[pro_id] = {
            "type": property_type,
            "label": label,
            "description": description,
        }
        return self

    def to_dict(self,):
        """Returns the query_object as a dict.

        Returns
        -------
        dict

        """
        return self.query_obj.to_dict()

    def to_json(self):
        """Returns the query_object as a json.

        Returns
        -------
        str

        """
        return self.query_obj.to_json()


class WOQLObj:
    def __init__(
        self,
        obj_id: str,
        obj_type: WOQLClass,
        label: str = None,
        description: str = None,
        obj_property: dict = None,
    ):
        """The WOQLObj constructor.

        Parameters
        ----------
        obj_id: str
            Object id for which the instance is created.
        obj_type: WOQLClass
            WOQLClass instance which has the properties.
        label: str, optional
            Optional label to be given for the instance.
        description: str, optional
            Optional description to be given for the instance.
        obj_property: dict, optional
            Optional properties which can be provided during initialization.

        Examples
        --------
        >>> my_id = "my_journey"
        >>> my_label = "My Journey"
        >>> my_des = "This is my journey to work"
        >>> my_prop = {"Duration": {"value": 30}}
        >>> journey_class = WOQLClass("Journey")
        >>> journey_class.add_property(
        ...     "Duration",
        ...     "integer",
        ...     label="Journey Duration",
        ...     description="Journey duration in minutes.",
        ... )
        <terminusdb_client.woqlquery.smart_query.WOQLClass object at 0x1085618d0>
        >>> woql_obj = WOQLObj(my_id, journey_class, my_label, my_des, obj_property=my_prop)
        """
        self.id = obj_id
        self._type = obj_type
        self.woql_id = self._idgen()
        self.label = label
        self.description = description
        self.query_obj = WOQLQuery().insert(
            self.woql_id, obj_type.id, label=label, description=description
        )

        if obj_property is not None:
            for pro_id, prop in obj_property.items():
                prop_val = prop.get("value")
                self._check_prop(pro_id, prop_val)
                self.query_obj = self.query_obj.property(
                    pro_id, prop_val, prop.get("label"), prop.get("description")
                )
            self._property = obj_property
        else:
            self._property = {}

    def __str__(self):
        """Returns the id of the instance

        Returns
        -------
        str
        """
        return self.id

    def _idgen(self) -> str:
        """Returns the id which is generated in the below format.
        doc:{[WOQLClass_type_id]_[WOQLObj_id]}

        Returns
        -------
        str

        """
        # mimic what a idgen would do in the back end
        # TODO: quote the ids ot make it url firendly
        return f"doc:{quote(self._type.id)}_{quote(self.id)}"

    def _check_prop(self, pro_id: str, pro_value):
        """Check if the given property's value belongs to the correct data type.

        Parameters
        ----------
        pro_id: str
            Property id
        pro_value
            Property value
        Returns
        -------
        none or raises exception

        """
        prop = self._type._property.get(pro_id)
        if prop is None:
            raise ValueError(f"No {pro_id} property in {self._type.id}")
        if isinstance(pro_value, WOQLObj):
            if pro_value._type != prop["type"]:
                raise ValueError(
                    f"{pro_id} property in {self._type.id} is of type {prop['type']} not {pro_value._type}"
                )
        else:
            if not isinstance(pro_value, WOQLTYPE_TO_PYTYPE[prop["type"]]):
                raise ValueError(
                    f"{pro_id} property in {self._type.id} is of type {prop['type']} not {type(pro_value)}"
                )

    def add_property(
        self, pro_id: str, pro_value, label: str = None, description: str = None
    ):
        """Adds the given properties to the instance.

        Parameters
        ----------
        pro_id: str
            Property id
        pro_value: any
            Property value which needs to be added.
        label: str, optional
            Optional label to be given for the instance.
        description: str, optional
            Optional description to be given for the instance.

        Returns
        -------
        WOQLObj

        """
        # check if the pro_value matches the property of the self._type
        self._check_prop(pro_id, pro_value)
        # add new prop in self._property
        self._property[pro_id] = {
            "value": pro_value,
            "label": label,
            "description": description,
        }
        # add to query_obj
        self.query_obj = self.query_obj.property(pro_id, pro_value, label, description)
        return self

    def to_dict(self,):
        """Returns the query_object as a dict.

        Returns
        -------
        dict

        """
        return self.query_obj.to_dict()

    def to_json(self):
        """Returns the query_object as a json.

        Returns
        -------
        str

        """
        return self.query_obj.to_json()


class TerminusDB:
    def __init__(
        self,
        server_url: str,
        db_id: str,
        key: str = "root",
        account: str = "admin",
        user: str = "admin",
        db_label: str = None,
        db_description: str = None,
        **kwargs,
    ):
        """The TerminusDB constructor.

        Parameters
        ----------
        server_url: str
            The url of which the TerminusDB server is running.
        db_id: str
            Unique identifier of the database.
        key: str, optional, default = "root"
        account: str, optional, default = "admin"
            ID of the organization in which to create the DB (defaults to 'admin')
        user: str, optional, default = "admin"
        db_label: str, optional, default = None
            Optional label to be given for the db.
        db_description: str, optional, default = None
            Optional description to be given for the db.
        **kwargs
            Configuration options used to construct a :class:`ConnectionConfig` instance.
            Passing insecure=True will skip HTTPS certificate checking.
        """

        from ..woqlclient import WOQLClient

        self._client = WOQLClient(server_url, **kwargs)
        self._client.connect(key=key, account=account, user=user)
        existing = self._client.get_database(db_id, self._client.uid())
        self.classes: Dict[str, WOQLClass] = {}
        if not existing:
            self._client.create_database(db_id, account, db_label, db_description)
        else:
            self._client.db(db_id)
            # get all classes from db and store them
            cls_result = WOQLLib().classes().execute(self._client)
            for item in cls_result["bindings"]:
                class_id = item["Class ID"].split("#")[-1]
                class_name = item["Class Name"]["@value"]
                class_des = item["Description"]["@value"]
                self.classes[class_id] = WOQLClass(class_id, class_name, class_des)
            # get all peoperties from db and add to classes
            prop_result = WOQLLib().property().execute(self._client)
            for item in prop_result["bindings"]:
                prop_domain = item["Property Domain"].split("#")[-1]
                prop_id = item["Property ID"].split("#")[-1]
                prop_name = item["Property Name"]["@value"]
                prop_des = item["Property Description"]["@value"]
                prop_type = item["Property Domain"].split("#")[-1]
                if item["Property Type"]["@value"] == "Object":
                    prop_type = self.classes[prop_type]
                self.classes[prop_domain].add_property(
                    prop_id, prop_type, prop_name, prop_des
                )

    def add_class(self, obj: Union[WOQLClass, List[WOQLClass]]):
        """Adds one or more WOQLClass types.

        Parameters
        ----------
        obj: WOQLClass or a list of WOQLClass instances
        """
        if isinstance(obj, WOQLClass):
            self.classes[obj.id] = obj
            return obj.query_obj.execute(self._client)
        elif isinstance(obj, list):
            for item in obj:
                self.classes[item.id] = item
            return WOQLQuery().woql_and(*obj).execute(self._client)
        else:
            raise ValueError(
                "object(s) added need to be WOQLClass object or a list of WOQLClass objects."
            )

    def add_object(self, obj: Union[WOQLObj, List[WOQLObj]]):
        """Adds one or more WOQLObj.

        Parameters
        ----------
        obj: WOQLObj or a list of WOQLObj instances

        Returns
        -------

        """
        if isinstance(obj, WOQLObj):
            # check if class is in db
            if obj._type.id in self.classes:
                return obj.query_obj.execute(self._client)
            else:
                raise ValueError("Class of object(s) is not in the schema.")
        elif isinstance(obj, list):
            for item in obj:
                self.classes[item.id] = item
            return WOQLQuery().woql_and(*obj).execute(self._client)
        else:
            raise ValueError(
                "Object(s) added need to be WOQLClass object or a list of WOQLClass objects."
            )

    def run(self, query: Union[WOQLQuery, Dict]):
        """Runs a query either in WOQLQuery format or json_ld in dictionary presentation

        Parameters
        ----------
        query: WOQLQuery or Dict
            Query which has to be executed.

        Returns
        -------
            The output of the query.

        """
        return self._client.query(query)



"""woqlClient.py"""
import copy
import json
import os
from typing import Any, BinaryIO, Dict, List, Optional, Tuple, Union

from ..__version__ import __version__
from ..woqlquery.woql_query import WOQLQuery
from .api_endpoint_const import APIEndpointConst
from .connectionCapabilities import ConnectionCapabilities

# from .errorMessage import *
from .connectionConfig import ConnectionConfig
from .dispatchRequest import DispatchRequest

# from .errors import (InvalidURIError)
# from .errors import doc, opts

# WOQL client object
# license Apache Version 2
# summary Python module for accessing the Terminus DB API


class WOQLClient:
    """Client for querying a TerminusDB server using WOQL queries."""

    def __init__(self, server_url: str, **kwargs) -> None:
        r"""The WOQLClient constructor.

        Parameters
        ----------
        server_url : str
            URL of the server that this client will connect to.
        \**kwargs
            Configuration options used to construct a :class:`ConnectionConfig` instance.
            Passing insecure=True will skip HTTPS certificate checking.
        """
        self.conConfig = ConnectionConfig(server_url, **kwargs)
        self.conCapabilities = ConnectionCapabilities()
        self.insecure = kwargs.get("insecure")

    def connect(self, **kwargs) -> dict:
        r"""Connect to a Terminus server at the given URI with an API key.

        Stores the terminus:ServerCapability document returned
        in the conCapabilities register which stores, the url, key, capabilities,
        and database meta-data for the connected server.

        If the ``serverURL`` argument is omitted,
        :attr:`self.conConfig.serverURL` will be used if present
        or an error will be raise.

        Parameters
        ----------
        \**kwargs
            Configuration options added to :attr:`conConfig`.

        Returns
        -------
        dict

        Examples
        -------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.connect(server_url="http://localhost:6363", key="<key>")
        dict
        """
        if len(kwargs) > 0:
            self.conConfig.update(**kwargs)
        if self.insecure is None:
            self.insecure = kwargs.get("insecure")

        json_obj = self.dispatch(APIEndpointConst.CONNECT, self.conConfig.api)
        self.conCapabilities.set_capabilities(json_obj)
        return json_obj

    def copy(self) -> "WOQLClient":
        """Create a deep copy of this client.

        Returns
        -------
        WOQLClient
            The copied client instance.

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> clone = client.copy()
        >>> assert client is not clone
        """
        return copy.deepcopy(self)

    def basic_auth(self, key: Optional[str] = None, user: Optional[str] = None):
        """Set or get the ``user:password`` for basic HTTP authentication to the server.

        If ``key`` is not provided, then the config will not be updated.

        Parameters
        ----------
        key : str, optional
            Optional password to use when authenticating to the server.
        user : str, optional
            Optional user name to use when authenticating to the server.

        Returns
        -------
        str:
            The basic authentication credentials in ``user:password`` format.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.basic_auth()
        False
        >>> client.basic_auth("password", "admin")
        'admin:password'
        >>> client.basic_auth()
        'admin:password'
        """
        if key:
            self.conConfig.set_basic_auth(key, user)
        return self.conConfig.basic_auth

    def remote_auth(self, auth_info: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """Set or get the Basic auth or JWT token used for authenticating to the server.

        If ``auth_info`` is not provided, then the config will not be updated.

        Parameters
        ----------
        auth_info : dict, optional
            Optional dict of authentication info containing
            ``"type"``, ``"user"`` and ``"key"`` keys.

        Returns
        -------
        dict
            The remote authentication info.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.remote_auth()
        >>> client.remote_auth({"type": "jwt", "user": "admin", "key": "<token>"})
        {'type': 'jwt', 'user': 'admin', 'key': '<token>'}
        """
        if auth_info:
            self.conConfig.set_remote_auth(auth_info)
        return self.conConfig.remote_auth

    def set_db(self, dbid: Optional[str]) -> str:
        """Set and return the current database.

        Parameters
        ----------
        dbid : str, optional
            Database identifer to set in the config.

        Returns
        -------
        str
            The current database identifier.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.set_db("<database>")
        '<database>'
        """
        return self.db(dbid)

    def db(self, dbid: Optional[str] = None) -> str:
        """Set or get the current database.

        If ``dbid`` is not provided, then the config will not be updated.

        Parameters
        ----------
        dbid : str, optional
            Optional database identifier to set in the config.

        Returns
        -------
        str
            The current database identifier.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.db()
        False
        >>> client.db("<database>")
        '<database>'
        >>> client.db()
        '<database>'
        """
        if dbid:
            self.conConfig.db = dbid
        return self.conConfig.db

    def account(self, accountid: Optional[str] = None) -> str:
        """Set or get the account identifier.

        If ``accountid`` is not provided, then the config will not be updated.

        Parameters
        ----------
        accountid : str, optional
            Optional account identifier to set in the config.

        Returns
        -------
        str
            The current account name.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.account()
        False
        >>> client.account("<account>")
        '<account>'
        >>> client.account()
        '<account>'
        """
        if accountid:
            self.conConfig.account = accountid
        return self.conConfig.account

    def user_account(self) -> str:
        """Get the current user identifier.

        Returns
        -------
        str
            User identifier.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.user_account()
        '<uid>'
        """
        u = self.conCapabilities.get_user()
        return u.id

    def user(self) -> Dict[str, Any]:
        """Get the current user's information.

        Returns
        -------
        dict

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.user()
        {'id': '<uid>', 'author': '<author>', 'roles': [], 'label': '<label>', 'comment': '<comment>'}
        """
        return self.conCapabilities.get_user()

    def repo(self, repoid: Optional[str] = None) -> str:
        """Set or get the repository identifier.

        If ``repoid`` is not provided, then the config will not be updated.

        Parameters
        ----------
        repoid : str, optional
            Optional repository identifier to set in the config.

        Returns
        -------
        str
            The current repository identifier.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.repo()
        'local'
        >>> client.repo("<repository>")
        '<repository>'
        >>> client.repo()
        '<repository>'
        """
        if repoid:
            self.conConfig.repo = repoid

        return self.conConfig.repo

    def ref(self, refid: Optional[str] = None) -> str:
        """Set or get the ref pointer (pointer to a commit within the branch).

        If ``refid`` is not provided, then the config will not be updated.

        Parameters
        ----------
        refid : str, optional
            Optional ref pointer to set in the config.

        Returns
        -------
        str
            The current ref pointer.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.ref()
        False
        >>> client.ref("<branch>")
        '<branch>'
        >>> client.ref()
        '<branch>'
        """
        if refid:
            self.conConfig.ref = refid
        return self.conConfig.ref

    def checkout(self, branchid: Optional[str] = None) -> str:
        """Set or get the current branch identifier.

        If ``branchid`` is not provided, then the config will not be updated.

        Parameters
        ----------
        branchid : str, optional
            Optional branch identifier to set in the config.

        Returns
        -------
        str
            The current branch identifier.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.checkout()
        'main'
        >>> client.checkout("<branch>")
        '<branch>'
        >>> client.checkout()
        '<branch>'
        """
        if branchid:
            self.conConfig.branch = branchid
        return self.conConfig.branch

    def uid(self, ignore_jwt: Optional[bool] = True) -> str:
        """Get the current user identifier.

        Parameters
        ----------
        ignore_jwt : bool, optional
            If ``True``, the local user identifier will be returned rather than the one set in the JWT.

        Returns
        -------
        str
            The local or JWT user identifier.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.uid()
        '<uid>'
        >>> client.uid(False)
        '<jwt_uid>'
        """
        return self.conConfig.user(ignore_jwt)

    def resource(self, ttype: str, val: Optional[str] = None) -> str:
        """Create a resource identifier string based on the current config.

        Parameters
        ----------
        ttype : str
            Type of resource. One of ["db", "meta", "repo", "commits", "ref", "branch"].
        val : str, optional
            Branch or commit identifier.

        Returns
        -------
        str
            The constructed resource string.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.account("<account>")
        '<account>'
        >>> client.db("<db>")
        '<db>'
        >>> client.repo("<repo>")
        '<repo>'
        >>> client.resource("db")
        '<account>/<db>/'
        >>> client.resource("meta")
        '<account>/<db>/_meta'
        >>> client.resource("commits")
        '<account>/<db>/<repo>/_commits'
        >>> client.resource("repo")
        '<account>/<db>/<repo>/_meta'
        >>> client.resource("ref", "<reference>")
        '<account>/<db>/<repo>/commit/<reference>'
        >>> client.resource("branch", "<branch>")
        '<account>/<db>/<repo>/branch/<branch>'
        """
        base = self.account() + "/" + self.db() + "/"
        if ttype == "db":
            return base
        elif ttype == "meta":
            return base + "_meta"
        base = base + self.repo()
        if ttype == "repo":
            return base + "/_meta"
        elif ttype == "commits":
            return base + "/_commits"
        if val is None:
            if ttype == "ref":
                val = self.ref()
            else:
                val = self.checkout()
        if ttype == "branch":
            return base + "/branch/" + val
        if ttype == "ref":
            return base + "/commit/" + val

    def set(self, **kwargs: Dict[str, Any]):  # bad naming
        r"""Update multiple config values on the current context.

        Parameters
        ----------
        \**kwargs
            Dict of config options to set.

        Examples
        --------
        >>> client = WOQLClient("http://localhost:6363")
        >>> client.set({"account": "<account>", "branch": "<branch>"})
        """
        self.conConfig.update(**kwargs)

    def create_database(
        self,
        dbid: str,
        accountid: Optional[str] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        prefixes: Optional[dict] = None,
        include_schema: bool = True,
    ) -> dict:
        """Create a TerminusDB database by posting
        a terminus:Database document to the Terminus Server.

        Parameters
        ----------
        dbid : str
            Unique identifier of the database.
        accountid : str, optional
            ID of the organization in which to create the DB (defaults to 'admin')
        label : str, optional
            Database name.
        description : str, optional
            Database description.
        prefixes : dict, optional
            Optional dict containing ``"doc"`` and ``"scm"`` keys.

            doc (str)
                IRI to use when ``doc:`` prefixes are expanded. Defaults to ``terminusdb:///data``.
            scm (str)
                IRI to use when ``scm:`` prefixes are expanded. Defaults to ``terminusdb:///schema``.
        include_schema : bool
            If ``True``, a main schema graph will be created, otherwise only a main instance graph will be created.

        Returns
        -------
        dict

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.create_database("someDB", "Database Label", "password")
        """
        details: Dict[str, Any] = {}
        if label:
            details["label"] = label
        else:
            details["label"] = dbid
        if description:
            details["comment"] = description
        else:
            details["comment"] = ""
        if include_schema:
            details["schema"] = True
        if prefixes:
            details["prefixes"] = prefixes
        if accountid is None:
            accountid = self.user_account()

        self.db(dbid)
        self.account(accountid)
        return self.dispatch(
            APIEndpointConst.CREATE_DATABASE, self.conConfig.db_url(), details
        )

    def delete_database(
        self, dbid: str, accountid: Optional[str] = None, force: bool = False
    ) -> dict:
        """Delete a TerminusDB database.

        If ``accountid`` is provided, then the account in the config will be updated
        and the new value will be used in future requests to the server.

        Parameters
        ----------
        dbid : str
            ID of the database to delete
        accountid : str, optional
            the account id in which the database resides (defaults to "admin")
        force: bool

        Returns
        -------
        dict

        Examples
        -------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.delete_database("<database>", "<account>")
        dict
        """

        self.db(dbid)
        if accountid:
            self.account(accountid)
        payload = {"force": force}
        json_response = self.dispatch(
            APIEndpointConst.DELETE_DATABASE, self.conConfig.db_url(), payload
        )
        return json_response

    def create_graph(self, graph_type: str, graph_id: str, commit_msg: str) -> dict:
        """Create a new graph in the current database context.

        Parameters
        ----------
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str
            Graph identifier.
        commit_msg : str
            Commit message.

        Returns
        -------
        dict

        Raises
        ------
        ValueError
            If the value of ``graph_type`` is invalid.
        """
        if graph_type in ["inference", "schema", "instance"]:
            commit = self._generate_commit(commit_msg)
            return self.dispatch(
                APIEndpointConst.CREATE_GRAPH,
                self.conConfig.graph_url(graph_type, graph_id),
                commit,
            )

        raise ValueError(
            "Create graph parameter error - you must specify a valid graph_type (inference, instance, schema), graph_id and commit message"
        )

    def delete_graph(self, graph_type: str, graph_id: str, commit_msg: str) -> dict:
        """Delete a graph from the current database context.

        Parameters
        ----------
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str
            Graph identifier.
        commit_msg : str
            Commit message.

        Returns
        -------
        dict

        Raises
        ------
        ValueError
            If the value of ``graph_type`` is invalid.
        """
        if graph_type in ["inference", "schema", "instance"]:
            commit = self._generate_commit(commit_msg)
            return self.dispatch(
                APIEndpointConst.DELETE_GRAPH,
                self.conConfig.graph_url(graph_type, graph_id),
                commit,
            )

        raise ValueError(
            "Delete graph parameter error - you must specify a valid graph_type (inference, instance, schema), graph_id and commit message"
        )

    def get_triples(self, graph_type: str, graph_id: str) -> dict:
        """Retrieves the contents of the specified graph as triples encoded in turtle format

        Parameters
        ----------
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str
            Graph identifier.

        Returns
        -------
        dict
        """
        return self.dispatch(
            APIEndpointConst.GET_TRIPLES,
            self.conConfig.triples_url(graph_type, graph_id),
        )

    def update_triples(
        self, graph_type: str, graph_id: str, turtle, commit_msg: str
    ) -> dict:
        """Updates the contents of the specified graph with the triples encoded in turtle format Replaces the entire graph contents

        Parameters
        ----------
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str
            Graph identifier.
        turtle
            Valid set of triples in Turtle format.
        commit_msg : str
            Commit message.

        Returns
        -------
        dict
        """
        commit = self._generate_commit(commit_msg)
        commit["turtle"] = turtle
        return self.dispatch(
            APIEndpointConst.UPDATE_TRIPLES,
            self.conConfig.triples_url(graph_type, graph_id),
            commit,
        )

    def insert_triples(
        self, graph_type: str, graph_id: str, turtle, commit_msg: str
    ) -> dict:
        """Inserts into the specified graph with the triples encoded in turtle format.

        Parameters
        ----------
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str
            Graph identifier.
        turtle
            Valid set of triples in Turtle format.
        commit_msg : str
            Commit message.

        Returns
        -------
        dict
        """
        commit = self._generate_commit(commit_msg)
        commit["turtle"] = turtle
        return self.dispatch(
            APIEndpointConst.INSERT_TRIPLES,
            self.conConfig.triples_url(graph_type, graph_id),
            commit,
        )

    def get_csv(
        self,
        csv_name: str,
        csv_directory: Optional[str] = None,
        graph_type: Optional[str] = None,
        graph_id: Optional[str] = None,
    ):
        """Retrieves the contents of the specified graph as a CSV

        Parameters
        ----------
        csv_name : str
            Name of csv to dump from the specified database to extract.
        csv_directory : str
            CSV output directory path. (defaults to current directory).
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str, optional
            Graph identifier.

        Returns
        -------
        dict
            An API success message
        """
        options = {}
        if csv_directory is None:
            csv_directory = os.getcwd()
        options["name"] = csv_name

        result = self.dispatch(
            APIEndpointConst.GET_CSV,
            self.conConfig.csv_url(graph_type, graph_id),
            options,
        )

        stream = open(f"{csv_directory}/{csv_name}", "w")
        stream.write(result.text)
        stream.close()
        return result

    def update_csv(
        self,
        csv_paths: Union[str, List[str]],
        commit_msg: str,
        graph_type: Optional[str] = None,
        graph_id: Optional[str] = None,
    ) -> dict:
        """Updates the contents of the specified graph with the triples encoded in turtle format Replaces the entire graph contents

        Parameters
        ----------
        csv_paths : str or list of str
            csv path or list of csv paths to load. (required)
        commit_msg : str
            Commit message.
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str, optional
            Graph identifier.

        Returns
        -------
        dict
            An API success message
        """
        if commit_msg is None:
            commit_msg = f"Update csv from {csv_paths} by python client {__version__}"
        commit = self._generate_commit(commit_msg)
        if isinstance(csv_paths, str):
            csv_paths_list = [csv_paths]
        else:
            csv_paths_list = csv_paths

        file_dict = {}
        for path in csv_paths_list:
            name = os.path.basename(os.path.normpath(path))
            file_dict[name] = (name, open(path, "rb"), "application/binary")

        return self.dispatch(
            APIEndpointConst.UPDATE_CSV,
            self.conConfig.csv_url(graph_type, graph_id),
            commit,
            file_dict=file_dict,
        )

    def insert_csv(
        self,
        csv_paths: Union[str, List[str]],
        commit_msg: str,
        graph_type: Optional[str] = None,
        graph_id: Optional[str] = None,
    ) -> dict:
        """Inserts into the specified graph with the triples encoded in turtle format.

        Parameters
        ----------
        csv_paths : str or list
            CSV or list of csvs to upload
        commit_msg : str
            Commit message.
        graph_type : str
            Graph type, either ``"inference"``, ``"instance"`` or ``"schema"``.
        graph_id : str
            Graph identifier.

        Returns
        -------
        dict
            An API success message
        """
        if commit_msg is None:
            commit_msg = f"Insert csv from {csv_paths} by python client {__version__}"
        commit = self._generate_commit(commit_msg)
        file_dict: Dict[str, Any] = {}
        if isinstance(csv_paths, str):
            csv_paths_list = [csv_paths]
        else:
            csv_paths_list = csv_paths

        file_dict = {}
        for path in csv_paths_list:
            name = os.path.basename(os.path.normpath(path))
            file_dict[name] = (name, open(path, "rb"), "application/binary")

        return self.dispatch(
            APIEndpointConst.INSERT_CSV,
            self.conConfig.csv_url(graph_type, graph_id),
            commit,
            file_dict=file_dict,
        )

    def query(
        self,
        woql_query: Union[dict, WOQLQuery],
        commit_msg: Optional[str] = None,
        file_dict: Optional[dict] = None,
    ):
        """Updates the contents of the specified graph with the triples encoded in turtle format Replaces the entire graph contents

        Parameters
        ----------
        woql_query : dict or WOQLQuery object
            A woql query as an object or dict
        commit_mg : str
            A message that will be written to the commit log to describe the change
        file_dict:
            File dictionary to be associated with post name => filename, for multipart POST

        Examples
        -------
        >>> WOQLClient(server="http://localhost:6363").query(woql, "updating graph")
        """
        if (
            hasattr(woql_query, "_contains_update_check")
            and woql_query._contains_update_check()  # type: ignore
        ):
            if not commit_msg:
                commit_msg = f"Update Query generated by python client {__version__}"
            query_obj = self._generate_commit(commit_msg)
        elif type(woql_query) is dict and commit_msg:
            query_obj = self._generate_commit(commit_msg)
        else:
            query_obj = {}

        if isinstance(woql_query, WOQLQuery):
            request_woql_query = woql_query.to_dict()
        else:
            request_woql_query = woql_query
        request_woql_query[
            "@context"
        ] = self.conCapabilities.get_context_for_outbound_query(None, self.db())
        query_obj["query"] = request_woql_query
        request_file_dict: Optional[Dict[str, Tuple[str, Union[str, BinaryIO], str]]]
        if file_dict is not None and type(file_dict) is dict:
            request_file_dict = {}
            for name in query_obj:
                query_obj_value = query_obj[name]
                request_file_dict[name] = (
                    name,
                    json.dumps(query_obj_value),
                    "application/json",
                )
            for name in file_dict:
                path = file_dict[name]
                request_file_dict[name] = (name, open(path, "rb"), "application/binary")
            payload = None
        else:
            request_file_dict = None
            payload = query_obj

        return self.dispatch(
            APIEndpointConst.WOQL_QUERY,
            self.conConfig.query_url(),
            payload,
            request_file_dict,
        )

    def branch(self, new_branch_id: str, empty: bool = False) -> dict:
        """Create a branch starting from the current branch.

        Parameters
        ----------
        new_branch_id : str
            New branch identifier.
        empty : bool
            Create an empty branch if true (no starting commit)

        Returns
        -------
        dict
        """
        if empty:
            source = {}
        elif self.ref():
            source = {
                "origin": f"{self.account()}/{self.db()}/{self.repo()}/commit/{self.ref()}"
            }
        else:
            source = {
                "origin": f"{self.account()}/{self.db()}/{self.repo()}/branch/{self.checkout()}"
            }

        return self.dispatch(
            APIEndpointConst.BRANCH, self.conConfig.branch_url(new_branch_id), source
        )

    def pull(self, remote_source_repo: dict) -> dict:
        """Pull updates from a remote repository to the current database.

        Parameters
        ----------
        remote_source_repo : dict
            Remote repository identifier containing ``"remote"`` and ``"remote_branch"`` keys.

        Returns
        -------
        dict

        Raises
        ------
        ValueError
            If the ``remote_source_repo`` is missing required keys.

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.pull({"remote": "<remote>", "remote_branch": "<branch>"})
        """
        rc_args = self._prepare_revision_control_args(remote_source_repo)
        if (
            rc_args is not None
            and rc_args.get("remote")
            and rc_args.get("remote_branch")
        ):
            return self.dispatch(
                APIEndpointConst.PULL, self.conConfig.pull_url(), rc_args,
            )
        else:
            raise ValueError(
                "Pull parameter error - you must specify a valid remote source and branch to pull from"
            )

    def fetch(self, remote_id: str):
        return self.dispatch(
            APIEndpointConst.FETCH, self.conConfig.fetch_url(remote_id)
        )

    def push(self, remote_target_repo: Dict[str, str]):
        """Push changes from a branch to a remote repo

        Parameters
        ----------
        remote_source_repo : dict
            details of the remote to push to [remote, remote_branch, message]

        Examples
        -------
        >>> WOQLClient(server="http://localhost:6363").push({remote: "origin", "remote_branch": "main", "author": "admin", "message": "message"})
        """
        rc_args = self._prepare_revision_control_args(remote_target_repo)
        if (
            rc_args is not None
            and rc_args.get("remote")
            and rc_args.get("remote_branch")
        ):
            return self.dispatch(
                APIEndpointConst.PUSH, self.conConfig.push_url(), rc_args
            )
        else:
            raise ValueError(
                "Push parameter error - you must specify a valid remote target"
            )

    def rebase(self, rebase_source: Dict[str, str]) -> dict:
        """Rebase the current branch onto the specified remote branch.

        Notes
        -----
        The "remote" repo can live in the local database.

        Parameters
        ----------
        rebase_source : dict
            Dict containing a ``"rebase_from"`` key.

        Returns
        -------
        dict

        Raises
        ------
        ValueError
            If the ``rebase_source`` is missing required keys.

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.rebase({"rebase_from": "<branch>"})
        """
        rc_args = self._prepare_revision_control_args(rebase_source)
        if rc_args is not None and rc_args.get("rebase_from"):
            return self.dispatch(
                APIEndpointConst.REBASE, self.conConfig.rebase_url(), rc_args
            )
        else:
            raise ValueError(
                "Rebase parameter error - you must specify a valid rebase source to rebase from"
            )

    def reset(self, commit_path: str) -> dict:
        """Reset the current branch HEAD to the specified commit path.

        Notes
        -----
        The "remote" repo can live in the local database.

        Parameters
        ----------
        commit_path : string
            Path to the commit, for instance admin/database/local/commit/234980523ffaf93.

        Returns
        -------
        dict

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.checkout("some_branch")
        >>> client.reset('admin/database/local/commit/234980523ffaf93')
        """

        return self.dispatch(
            APIEndpointConst.RESET,
            self.conConfig.reset_url(),
            {"commit_descriptor": commit_path},
        )

    def optimize(self, path: str) -> dict:
        """Optimize the specified path.

        Notes
        -----
        The "remote" repo can live in the local database.

        Parameters
        ----------
        path : string
            Path to optimize, for instance admin/database/_meta for the repo graph.

        Returns
        -------
        dict

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.optimize('admin/database/_meta')
        """
        return self.dispatch(APIEndpointConst.RESET, self.conConfig.optimize_url(path))

    def squash(self, msg: str, author: Optional[str] = None) -> dict:
        """Squash the current branch HEAD into a commit

        Notes
        -----
        The "remote" repo can live in the local database.

        Parameters
        ----------
        msg : string
            Message for the newly created squash commit
        author : string
            Author of the commit

        Returns
        -------
        dict
            A dict with the new commit id:
            {'@type' : 'api:SquashResponse',
             'api:commit' : Commit,
             'api:old_commit' : Old_Commit,
             'api:status' : "api:success"}

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.connect(user="admin", key="root", account="admin", db="some_db")
        >>> client.squash('This is a squash commit message!')
        """
        commit_object = self._generate_commit(msg, author)
        return self.dispatch(
            APIEndpointConst.SQUASH, self.conConfig.squash_url(), commit_object
        )

    def clonedb(self, clone_source: Dict[str, str], newid: str) -> dict:
        """Clone a remote repository and create a local copy.

        Parameters
        ----------
        clone_source : dict
            Dict containing a ``"remote_url"`` key.
        newid : str
            Identifier of the new repository to create.

        Returns
        -------
        dict

        Raises
        ------
        ValueError
            If the ``clone_source`` is missing required keys.

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client.clonedb({"remote_url": "<remote_url>"}, "<newid>")
        """
        rc_args = self._prepare_revision_control_args(clone_source)
        if rc_args is not None and rc_args.get("remote_url"):
            return self.dispatch(
                APIEndpointConst.CLONE, self.conConfig.clone_url(newid), rc_args
            )
        else:
            raise ValueError(
                "Clone parameter error - you must specify a valid id for the cloned database"
            )

    def _generate_commit(self, msg: str, author: Optional[str] = None) -> dict:
        """Pack the specified commit info into a dict format expected by the server.

        Parameters
        ----------
        msg : str
            Commit message.
        author : str
            Commit author.

        Returns
        -------
        dict
            Formatted commit info.

        Examples
        --------
        >>> client = WOQLClient("https://127.0.0.1:6363/")
        >>> client._generate_commit("<message>", "<author>")
        {'commit_info': {'author': '<author>', 'message': '<message>'}}
        """
        if author:
            mes_author = author
        else:
            mes_author = self.conCapabilities.author()

        ci = {"commit_info": {"author": mes_author, "message": msg}}
        return ci

    def _prepare_revision_control_args(
        self, rc_args: Optional[dict] = None
    ) -> Optional[dict]:
        """Ensure the ``"author"`` field in the specified argument dict is set.
        If ``"author"`` is not in ``rc_args``, the current author value will be set.

        Parameters
        ----------
        rc_args : dict, optional
            Optional dict containing arguments used in revision control actions.

        Returns
        -------
        dict, optional
            ``None`` if ``rc_args`` is not provided, otherwise the modified ``rc_args``.
        """
        if rc_args is None:
            return None
        if not rc_args.get("author"):
            rc_args["author"] = self.conCapabilities.author()
        return rc_args

    def dispatch(
        self,
        action: Union[str, Tuple[str]],
        url: str,
        payload: Optional[dict] = None,
        file_dict: Optional[dict] = None,
    ) -> dict:
        """Directly dispatch to a TerminusDB database.

        Parameters
        ----------
        action
            The action to perform on the server.
        url : str
            The server URL to point the action at.
        payload : dict, optional
            Payload to send to the server.
        file_dict : dict, optional
            Dict of files to include in the query.

        Returns
        -------
        dict
        """

        # check if we can perform this action or raise an AccessDeniedError error
        # review the access control
        # self.conCapabilities.capabilitiesPermit(action)
        # url, action, payload={}, basic_auth, jwt=None, file_dict=None)

        if payload is None:
            payload = {}

        return DispatchRequest.send_request_by_action(
            url,
            action,
            payload,
            self.basic_auth(),
            self.remote_auth(),
            file_dict,
            self.insecure,
        )

    def get_database(self, dbid: str, account: str) -> Optional[dict]:
        """
        Returns metadata (id, organization, label, comment) about the requested database
        Parameters
        ----------
        dbid : str
            The id of the database
        account : str
            The account / organization id that the user is acting through

        Returns
        -------
        dict or None if not found
        """
        return self.conCapabilities.get_database(dbid, account)

    def get_databases(self) -> List[Dict]:
        """
        Returns a list of database metadata records for all databases the user has access to

        Returns
        -------
        list of dicts
        """
        return self.conCapabilities.get_databases()

    def get_metadata(self, dbid: str, account):
        """
        Alias of get_database above - deprecated - included for backwards compatibility
        """
        return self.get_database(dbid, account)

    def server(self) -> str:
        """
        Returns the URL of the currently connected server
        """
        return self.conConfig.server

    def api(self) -> str:
        """
        Returns the URL of the currently connected server
        """
        return self.conConfig.api

    """
    Unstable / Experimental Endpoints

    The below API endpoints are not yet officially released

    They are part of the server API (not desktop) and will be finalized and
    officially released when that package is released. They should be considered
    unreliable and unstable until then and use is unsupported and at your own risk
    """

    def get_class_frame(self, class_name):
        opts = {"class": class_name}
        return self.dispatch(
            APIEndpointConst.CLASS_FRAME, self.conConfig.class_frame_url(), opts,
        )



from .woqlclient import WOQLClient  # noqa
from .woqldataframe import woqlDataframe as WOQLDataFrame  # noqa
from .woqlquery import TerminusDB  # noqa
from .woqlquery import WOQLClass  # noqa
from .woqlquery import WOQLLib  # noqa
from .woqlquery import WOQLObj  # noqa
from .woqlquery import WOQLQuery  # noqa
from .woqlview import WOQLView  # noqa



import os
import subprocess
import time
import unittest

import pytest
import requests
from terminusdb_client.woqlquery.smart_query import (
    TerminusDB,
    WOQLClass,
    WOQLLib,
    WOQLObj,
)

MAX_CONTAINER_STARTUP_TIME = 30


def is_docker_installed():
    try:
        output = subprocess.run(["docker", "--version"], stderr=subprocess.PIPE)
    except FileNotFoundError:
        return False
    return output.returncode == 0


pytestmark = pytest.mark.skipif(
    not is_docker_installed(), reason="docker not installed"
)


@pytest.fixture(scope="module")
def docker_url(pytestconfig):
    # we are using subprocess in case we need to access some of the outputs
    # most likely
    pytestconfig.getoption("docker_compose")
    output = subprocess.run(
        [
            "docker-compose",
            "--file",
            os.path.dirname(os.path.realpath(__file__)) + "/test-docker-compose.yml",
            "up",
            "-d",
        ],
        stderr=subprocess.PIPE,
    )
    if output.returncode != 0:
        raise RuntimeError(output.stderr)

    test_url = "http://127.0.0.1:6363"
    is_server_started = False

    seconds_waited = 0
    while not is_server_started:
        service = subprocess.run(
            [
                "docker-compose",
                "--file",
                os.path.dirname(os.path.realpath(__file__))
                + "/test-docker-compose.yml",
                "ps",
                "--services",
                "--filter",
                "status=running",
            ],
            stdout=subprocess.PIPE,
            check=True,
        )

        if service.stdout == b"terminusdb-server\n":
            try:
                response = requests.get(test_url)
                assert response.status_code == 200
                break
            except (requests.exceptions.ConnectionError, AssertionError):
                pass

        seconds_waited += 1
        time.sleep(1)

        if seconds_waited > MAX_CONTAINER_STARTUP_TIME:
            clean_up_container()
            raise RuntimeError("Container was to slow to startup")

    yield test_url
    clean_up_container()


def clean_up_container():
    subprocess.run(
        [
            "docker-compose",
            "--file",
            os.path.dirname(os.path.realpath(__file__)) + "/test-docker-compose.yml",
            "down",
        ],
        check=True,
    )
    subprocess.run(
        [
            "docker-compose",
            "--file",
            os.path.dirname(os.path.realpath(__file__)) + "/test-docker-compose.yml",
            "rm",
            "--force",
            "--stop",
            "-v",
        ],
        check=True,
    )
    subprocess.run(["docker-compose", "down"])
    subprocess.run(["docker-compose", "rm", "--force", "--stop", "-v"])


def test_main_service_run(docker_url):
    result = requests.get(docker_url)
    assert result.status_code == 200


def test_init_terminusdb(docker_url):
    db = TerminusDB(docker_url, "test")
    assert db._client.db("test") == "test"


def test_add_class_and_object(
    docker_url, one_class_obj, one_class_prop, one_object, one_prop_val
):
    db = TerminusDB(docker_url, "test")
    my_class = WOQLClass(
        "Journey",
        label="Bike Journey",
        description="Bike Journey object that capture each bike joourney.",
    )
    my_class.add_property("Duration", "integer")
    db.add_class(my_class)

    case = unittest.TestCase()
    case.assertCountEqual(db.run(WOQLLib().classes()), one_class_obj)
    case.assertCountEqual(db.run(WOQLLib().property()), one_class_prop)
    # assert db.run(WOQLLib().classes()) == one_class_obj
    # assert db.run(WOQLLib().property()) == one_class_prop

    my_obj = WOQLObj("myobj", my_class)
    my_obj.add_property("Duration", 75)
    db.add_object(my_obj)

    case.assertCountEqual(db.run(WOQLLib().objects()), one_object)
    case.assertCountEqual(db.run(WOQLLib().property_values()), one_prop_val)
    # assert db.run(WOQLLib().objects()) == one_object
    # assert db.run(WOQLLib().property_values()) == one_prop_val



# import sys
# sys.path.append('woqlclient')
import unittest.mock as mock

import pytest
import requests
from terminusdb_client.woqlclient.woqlClient import WOQLClient

from .mockResponse import mocked_requests
from .woqljson.woqlStarJson import WoqlStar


def mock_func_with_1arg(_):
    return True


def mock_func_with_2arg(first, second):
    return True


def mock_func_no_arg():
    return True


@mock.patch("requests.get", side_effect=mocked_requests)
def test_connection(mocked_requests, monkeypatch):

    woql_client = WOQLClient("http://localhost:6363")

    # before connect it connection is empty

    woql_client.connect(key="root", account="admin", user="admin")

    requests.get.assert_called_once_with(
        "http://localhost:6363/api/",
        headers={"Authorization": "Basic YWRtaW46cm9vdA=="},
        verify=False,
    )


@mock.patch("requests.post", side_effect=mocked_requests)
@mock.patch("requests.get", side_effect=mocked_requests)
def test_create_database(mocked_requests, mocked_requests2):
    woql_client = WOQLClient(
        "http://localhost:6363", user="admin", key="root", account="admin"
    )
    woql_client.connect()
    assert woql_client.basic_auth() == "admin:root"

    woql_client.create_database(
        "myFirstTerminusDB",
        "admin",
        label="my first db",
        description="my first db comment",
        include_schema=False,
    )

    requests.post.assert_called_once_with(
        "http://localhost:6363/api/db/admin/myFirstTerminusDB",
        headers={
            "Authorization": "Basic YWRtaW46cm9vdA==",
            "content-type": "application/json",
        },
        verify=False,
        json={"label": "my first db", "comment": "my first db comment"},
    )


@mock.patch("requests.post", side_effect=mocked_requests)
@mock.patch("requests.get", side_effect=mocked_requests)
@mock.patch("terminusdb_client.woqlclient.woqlClient.WOQLClient.create_graph")
def test_create_database_with_schema(
    mocked_requests, mocked_requests2, create_schema_obj
):
    woql_client = WOQLClient(
        "http://localhost:6363", user="admin", account="admin", key="root"
    )
    woql_client.connect()
    assert woql_client.basic_auth() == "admin:root"

    woql_client.create_database(
        "myFirstTerminusDB",
        "admin",
        label="my first db",
        description="my first db comment",
        include_schema=True,
    )

    requests.post.assert_called_once_with(
        "http://localhost:6363/api/db/admin/myFirstTerminusDB",
        headers={
            "Authorization": "Basic YWRtaW46cm9vdA==",
            "content-type": "application/json",
        },
        verify=False,
        json={"label": "my first db", "comment": "my first db comment", "schema": True},
    )


@mock.patch("requests.post", side_effect=mocked_requests)
@mock.patch("requests.get", side_effect=mocked_requests)
def test_create_database_and_change_account(mocked_requests, mocked_requests2):
    woql_client = WOQLClient(
        "http://localhost:6363", user="admin", account="admin", key="root"
    )
    woql_client.connect()
    woql_client.create_database(
        "myFirstTerminusDB",
        "my_new_account",
        label="my first db",
        description="my first db comment",
        include_schema=False,
    )

    requests.post.assert_called_once_with(
        "http://localhost:6363/api/db/my_new_account/myFirstTerminusDB",
        headers={
            "Authorization": "Basic YWRtaW46cm9vdA==",
            "content-type": "application/json",
        },
        verify=False,
        json={"label": "my first db", "comment": "my first db comment"},
    )

    assert woql_client.basic_auth() == "admin:root"


@mock.patch("requests.get", side_effect=mocked_requests)
@mock.patch("requests.post", side_effect=mocked_requests)
def test_branch(mocked_requests, mocked_requests2):
    woql_client = WOQLClient("http://localhost:6363")
    woql_client.connect(user="admin", account="admin", key="root", db="myDBName")
    woql_client.branch("my_new_branch")

    requests.post.assert_called_once_with(
        "http://localhost:6363/api/branch/admin/myDBName/local/branch/my_new_branch",
        headers={
            "Authorization": "Basic YWRtaW46cm9vdA==",
            "content-type": "application/json",
        },
        verify=False,
        json={"origin": "admin/myDBName/local/branch/main"},
    )


@mock.patch("requests.get", side_effect=mocked_requests)
@mock.patch("requests.post", side_effect=mocked_requests)
def test_wrong_graph_type(mocked_requests, mocked_requests2):
    woql_client = WOQLClient("http://localhost:6363")
    woql_client.connect(user="admin", account="admin", key="root", db="myDBName")

    with pytest.raises(ValueError):
        woql_client.create_graph("wrong_graph_name", "mygraph", "add a new graph")


@mock.patch("requests.get", side_effect=mocked_requests)
def test_get_triples(mocked_requests):
    woql_client = WOQLClient("http://localhost:6363")
    woql_client.connect(user="admin", account="admin", key="root", db="myDBName")

    woql_client.get_triples("instance", "mygraph")

    requests.get.assert_called_with(
        "http://localhost:6363/api/triples/admin/myDBName/local/branch/main/instance/mygraph",
        headers={"Authorization": "Basic YWRtaW46cm9vdA=="},
        verify=False,
    )


@mock.patch("requests.post", side_effect=mocked_requests)
@mock.patch("requests.get", side_effect=mocked_requests)
def test_query(mocked_requests, mocked_requests2):
    woql_client = WOQLClient("http://localhost:6363")
    woql_client.connect(user="admin", account="admin", key="root", db="myDBName")

    # WoqlStar is the query in json-ld

    woql_client.query(WoqlStar)

    requests.post.assert_called_once_with(
        "http://localhost:6363/api/woql/admin/myDBName/local/branch/main",
        headers={
            "Authorization": "Basic YWRtaW46cm9vdA==",
            "content-type": "application/json",
        },
        verify=False,
        json={"query": WoqlStar},
    )



ConnectResponse = {
    "@context": {
        "doc": "terminusdb:///system/data/",
        "layer": "http://terminusdb.com/schema/layer#",
        "owl": "http://www.w3.org/2002/07/owl#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "ref": "http://terminusdb.com/schema/ref#",
        "repo": "http://terminusdb.com/schema/repository#",
        "system": "http://terminusdb.com/schema/system#",
        "terminus": "http://terminusdb.com/schema/system#",
        "vio": "http://terminusdb.com/schema/vio#",
        "woql": "http://terminusdb.com/schema/woql#",
        "xdd": "http://terminusdb.com/schema/xdd#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    },
    "@id": "doc:anonymous",
    "@type": "system:User",
    "rdfs:comment": {"@language": "en", "@value": "This is the anonymous account"},
    "rdfs:label": {"@language": "en", "@value": "Anonymous User"},
    "system:agent_name": {"@type": "xsd:string", "@value": "anonymous"},
    "system:role": {
        "@id": "doc:anonymous_role",
        "@type": "system:Role",
        "rdfs:comment": {
            "@language": "en",
            "@value": "Role providing public access capabilities",
        },
        "rdfs:label": {"@language": "en", "@value": "Anonymous Role"},
    },
    "system:user_key_hash": "",
}



import json
import warnings
from numbers import Number

try:
    from IPython.display import Javascript, display
except ImportError:
    msg = "woqlview need to be used in Jupyter notebook.\n"
    warnings.warn(msg)


class WOQLView:
    def __init__(self):
        self.config = ""
        self.obj = None
        display(
            Javascript(
                """
        require.config({
            paths: {
                TerminusClient:'https://unpkg.com/@terminusdb/terminusdb-client@3.0.3/dist/terminusdb-client.min',
                TerminusDBGraph:'https://dl.bintray.com/terminusdb/terminusdb/dev/terminusdb-d3-graph.min'
            }
        });
        """
            )
        )

    def edges(self, *args: list):
        for item in args:
            if not isinstance(item, list):
                raise TypeError("argument of edges need to be lists")
        arguments = ",".join(map(str, args))
        self.config += f"woqlGraphConfig.edges({arguments});\n"
        return self

    def height(self, height_input: Number):
        self.config += f"woqlGraphConfig.height({height_input});\n"
        return self

    def width(self, width_input: Number):
        self.config += f"woqlGraphConfig.width({width_input});\n"
        return self

    def edge(self, start: str, end: str):
        if not isinstance(start, str) or not isinstance(end, str):
            raise TypeError("arguments of edge() need to be strings")
        arguments = ",".join(map(lambda x: f'"{x}"', [start, end]))
        self.obj = f"woqlGraphConfig.edge({arguments})"
        return self

    def node(self, *args: str):
        # if not isinstance(select_node, str):
        #     raise TypeError("argument of node() need to be string")
        # self.obj = f'woqlGraphConfig.node("{select_node}")'
        # return self

        for item in args:
            if not isinstance(item, str):
                raise TypeError("arguments of node() need to be strings")
        arguments = ",".join(map(lambda x: f'"{x}"', args))
        self.obj = f"woqlGraphConfig.node({arguments})"
        return self

    def text(self, input_text: str):
        if self.obj is None:
            raise SyntaxError("text() should be used following a node() or edge()")
        self.config += self.obj + f'.text("{input_text}");\n'
        return self

    def distance(self, input_distance: Number):
        if self.obj is None:
            raise SyntaxError("distance() should be used following a node() or edge()")
        self.config += self.obj + f".distance({input_distance});\n"
        return self

    def weight(self, input_weight: Number):
        if self.obj is None:
            raise SyntaxError("weight() should be used following a node() or edge()")
        self.config += self.obj + f".weight({input_weight});\n"
        return self

    def color(self, input_color: list):
        if not isinstance(input_color, list):
            raise TypeError("argument of color() need to be list")
        if len(input_color) > 3:
            input_color = input_color[:3]
        if len(input_color) < 3:
            input_color += [0] * (3 - len(input_color))
        if self.obj is None:
            raise SyntaxError("color() should be used following a node() or edge()")
        arguments = ",".join(map(str, input_color))
        self.config += self.obj + f".color([{arguments}]);\n"
        return self

    def icon(self, input_dict: dict):
        if not isinstance(input_dict, dict):
            raise TypeError("argument of icon() need to be dict")
        if self.obj is None:
            raise SyntaxError("icon() should be used following a node() or edge()")
        self.config += self.obj + f".icon({json.dumps(input_dict)});\n"
        return self

    def size(self, input_size: Number):
        if self.obj is None:
            raise SyntaxError("size() should be used following a node() or edge()")
        self.config += self.obj + f".size({input_size});\n"
        return self

    def collision_radius(self, input_radius: Number):
        if self.obj is None:
            raise SyntaxError(
                "collision_radius() should be used following a node() or edge()"
            )
        self.config += self.obj + f".collisionRadius({input_radius});\n"
        return self

    def hidden(self, input_choice: bool):
        if input_choice:
            self.config += self.obj + ".hidden(true);\n"
        else:
            self.config += self.obj + ".hidden(false);\n"
        return self

    def charge(self, input_charge: Number):
        if self.obj is None:
            raise SyntaxError("charge(() should be used following a node() or edge()")
        self.config += self.obj + f".charge({input_charge});\n"
        return self

    def of(self, input_obj: str):
        if self.obj is None:
            raise SyntaxError("in() should be used following a node() or edge()")
        self.obj += f'.in("{input_obj}")'
        return self

    def show(self, result: dict):
        """Show the graph inline in the Jupyter notebook

        Parameter
        ---------
        result: the result that is returning from a query in dict format."""
        display(
            Javascript(
                """
        (function(element){
        require(['TerminusClient','TerminusDBGraph'], function(TerminusClient,TerminusDBGraph){

            console.log(TerminusDBGraph);
            const resultData=%s

            const woqlGraphConfig= TerminusClient.View.graph();
            woqlGraphConfig.height(window.innerHeight).width(window.innerWidth);
            %s


            var result = new TerminusClient.WOQLResult(resultData);
            let viewer = woqlGraphConfig.create(null);

            viewer.setResult(result);
            const graphResult= new TerminusDBGraph.GraphResultsViewer(viewer.config, viewer);
            graphResult.load(element.get(0),true);
            })
            })(element)
            """
                % (result, self.config)
            )
        )

    def export(self, filename: str, result: dict):
        """Export the graph into an html file

        Parameter
        ---------
        filename: the file name of the export file (without extention).

        result: the result that is returning from a query in dict format."""
        with open(filename + ".html", "w", encoding="utf-8") as file:
            file.write(
                """<html>
              <head>
                <meta name="viewport" content="width=device-width, initial-scale=1">

                <link rel="stylesheet" href="https://use.fontawesome.com/releases/v5.0.9/css/all.css" integrity="sha384-5SOiIsAziJl6AWe0HWRKTXlfcSHKmYV4RBF18PPJ173Kzn7jzMyFuTtk8JA7QQG1" crossorigin="anonymous">

                <title>%s</title>
              </head>
              <body>
                 <div style="width: 100vw; height: 100vw;" id="mycontainer" ></div>
              </body>

              <script src="https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.6/require.js" integrity="sha512-vRqhAr2wsn+/cSsyz80psBbCcqzz2GTuhGk3bq3dAyytz4J/8XwFqMjiAGFBj+WM95lHBJ9cDf87T3P8yMrY7A==" crossorigin="anonymous"></script>

              <script>

              require.config({
                paths: {
                    TerminusClient:'https://unpkg.com/@terminusdb/terminusdb-client@3.0.3/dist/terminusdb-client.min',
                    TerminusDBGraph:'https://dl.bintray.com/terminusdb/terminusdb/dev/terminusdb-d3-graph.min'
                }
              });

                require(['TerminusClient','TerminusDBGraph'], function(TerminusClient,TerminusDBGraph){


                const resultData=%s

                const woqlGraphConfig= new TerminusClient.View.graph();
                woqlGraphConfig.height(500).width(800);
                %s

                var result = new TerminusClient.WOQLResult(resultData);
                let viewer = woqlGraphConfig.create(null);

                viewer.setResult(result);
                const graphResult= new TerminusDBGraph.GraphResultsViewer(viewer.config, viewer);
                graphResult.load("#mycontainer",true);

              })

              </script>


            </html>"""
                % (filename, result, self.config)
            )

    def print_js_config(self):
        """Print out the JavaScript config

        Parameter
        ---------
        result: the result that is returning from a query in dict format."""
        print(self.config)  # noqa



from datetime import datetime

from terminusdb_client.woqlquery.smart_query import WOQLClass, WOQLObj

from .ans_doctype import *  # noqa
from .ans_property import *  # noqa


class TestWOQLClass:
    def test_init(
        self,
        doctype_without,
        doctype_with_label,
        doctype_with_des,
        property_without,
        property_with_des,
    ):
        woql_object = WOQLClass("Station")
        woql_object_label = WOQLClass("Station", label="Station Object")
        woql_object_des = WOQLClass(
            "Station", label="Station Object", description="A bike station object."
        )
        woql_object_prop = WOQLClass(
            "Journey", obj_property={"Duration": {"type": "dateTime"}}
        )
        woql_object_prop_des = WOQLClass(
            "Journey",
            obj_property={
                "Duration": {
                    "type": "dateTime",
                    "label": "Journey Duration",
                    "description": "Journey duration in minutes.",
                }
            },
        )
        assert woql_object.to_dict() == doctype_without
        assert woql_object_label.to_dict() == doctype_with_label
        assert woql_object_des.to_dict() == doctype_with_des
        assert woql_object_prop.to_dict() == property_without
        assert woql_object_prop_des.to_dict() == property_with_des

    def test_label(self, doctype_with_label):
        woql_object = WOQLClass("Station")
        assert woql_object.label == ""
        woql_object.label = "Station Object"
        assert woql_object.label == "Station Object"
        assert woql_object.to_dict() == doctype_with_label

    def test_description(self, doctype_with_des):
        woql_object = WOQLClass("Station", label="Station Object")
        assert woql_object.description == ""
        woql_object.description = "A bike station object."
        assert woql_object.description == "A bike station object."
        assert woql_object.to_dict() == doctype_with_des

    def test_add_datatype_property(self, property_without, property_with_des):
        woql_object = WOQLClass("Journey")
        woql_object.add_property("Duration", "dateTime")
        assert woql_object.to_dict() == property_without

        woql_object = WOQLClass("Journey")
        woql_object.add_property(
            "Duration",
            "dateTime",
            label="Journey Duration",
            description="Journey duration in minutes.",
        )
        assert woql_object.to_dict() == property_with_des

    def test_add_object_property(self, obj_property_without):
        station_obj = WOQLClass("Station")
        woql_object = WOQLClass("Journey")
        woql_object.add_property("start_station", station_obj)
        assert woql_object.to_dict() == obj_property_without


class TestWOQLObj:
    def test_init_int_prop(self):
        my_id = "my_journey"
        my_label = "My Journey"
        my_des = "This is my journey to work"
        my_prop = {"Duration": {"value": 30}}
        journey_class = WOQLClass("Journey")
        journey_class.add_property(
            "Duration",
            "integer",
            label="Journey Duration",
            description="Journey duration in minutes.",
        )
        woql_obj = WOQLObj(my_id, journey_class, my_label, my_des, obj_property=my_prop)

        assert woql_obj.id == my_id
        assert woql_obj.label == my_label
        assert woql_obj.description == my_des
        assert woql_obj._type == journey_class
        assert woql_obj._property == my_prop

        def test_init_dt_prop(self):
            my_id = "my_journey"
            my_label = "My Journey"
            my_des = "This is my journey to work"
            my_prop = {"Duration": {"value": datetime("2020-08-08")}}
            journey_class = WOQLClass("Journey")
            journey_class.add_property(
                "Duration",
                "dateTime",
                label="Journey Duration",
                description="Journey duration in minutes.",
            )
            woql_obj = WOQLObj(
                my_id, journey_class, my_label, my_des, obj_property=my_prop
            )

            assert woql_obj.id == my_id
            assert woql_obj.label == my_label
            assert woql_obj.description == my_des
            assert woql_obj._type == journey_class
            assert woql_obj._property == my_prop

    def test_idgen(self):
        journey_class = WOQLClass("Journey")
        woql_obj = WOQLObj("my_journey", journey_class)
        assert woql_obj.woql_id == "doc:Journey_my_journey"

    def test_idgen_quote(self):
        journey_class = WOQLClass("My Journey")
        woql_obj = WOQLObj("my journey", journey_class)
        assert woql_obj.woql_id == "doc:My%20Journey_my%20journey"
